#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='namd_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='namd.nomadmetainfo.json'))


class x_namd_mdin_input_output_files(MCategory):
    '''
    Parameters of mdin belonging to x_namd_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_namd_mdin_input_output_files'))


class x_namd_mdin_control_parameters(MCategory):
    '''
    Parameters of mdin belonging to x_namd_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_namd_mdin_control_parameters'))


class x_namd_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_namd_mdin_method'))


class x_namd_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_namd_mdout_single_configuration_calculation'))


class x_namd_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_namd_mdout_method'))


class x_namd_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_namd_mdout_run'))


class x_namd_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_namd_mdin_run'))


class x_namd_section_input_output_files(MSection):
    '''
    Section to store input and output file names
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_namd_section_input_output_files'))

    x_namd_inout_file_structure = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input topology file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_structure'))

    x_namd_inout_file_traj_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output trajectory file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_traj_coord'))

    x_namd_inout_file_traj_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output file for velocities in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_traj_vel'))

    x_namd_inout_file_traj_force = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output file for forces in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_traj_force'))

    x_namd_inout_file_output_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_output_coord'))

    x_namd_inout_file_output_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_output_vel'))

    x_namd_inout_file_output_force = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD output forces file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_output_force'))

    x_namd_inout_file_input_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_input_coord'))

    x_namd_inout_file_input_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD input velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_input_vel'))

    x_namd_inout_file_restart_coord = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD restart coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_restart_coord'))

    x_namd_inout_file_restart_vel = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD restart velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_restart_vel'))

    x_namd_inout_file_fftw_datafile = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD FFTW data file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_fftw_datafile'))

    x_namd_inout_file_mdlog = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD MD output log file.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_file_mdlog'))


class x_namd_section_control_parameters(MSection):
    '''
    Section to store the input and output control parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_namd_section_control_parameters'))

    x_namd_inout_control_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_timestep'))

    x_namd_inout_control_number_of_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_number_of_steps'))

    x_namd_inout_control_steps_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_steps_per_cycle'))

    x_namd_inout_control_periodic_cell_basis_1 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_periodic_cell_basis_1'))

    x_namd_inout_control_periodic_cell_basis_2 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_periodic_cell_basis_2'))

    x_namd_inout_control_periodic_cell_basis_3 = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_periodic_cell_basis_3'))

    x_namd_inout_control_periodic_cell_center = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_periodic_cell_center'))

    x_namd_inout_control_load_balancer = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_load_balancer'))

    x_namd_inout_control_load_balancing_strategy = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_load_balancing_strategy'))

    x_namd_inout_control_ldb_period = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_ldb_period'))

    x_namd_inout_control_first_ldb_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_first_ldb_timestep'))

    x_namd_inout_control_last_ldb_timestep = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_last_ldb_timestep'))

    x_namd_inout_control_ldb_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_ldb_background_scaling'))

    x_namd_inout_control_hom_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_hom_background_scaling'))

    x_namd_inout_control_pme_background_scaling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_background_scaling'))

    x_namd_inout_control_min_atoms_per_patch = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_min_atoms_per_patch'))

    x_namd_inout_control_initial_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_initial_temperature'))

    x_namd_inout_control_center_of_mass_moving_initially = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_center_of_mass_moving_initially'))

    x_namd_inout_control_dielectric = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_dielectric'))

    x_namd_inout_control_excluded_species_or_groups = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_excluded_species_or_groups'))

    x_namd_inout_control_1_4_electrostatics_scale = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_1_4_electrostatics_scale'))

    x_namd_inout_control_traj_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_traj_dcd_filename'))

    x_namd_inout_control_traj_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_traj_dcd_frequency'))

    x_namd_inout_control_traj_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_traj_dcd_first_step'))

    x_namd_inout_control_velocity_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_dcd_filename'))

    x_namd_inout_control_velocity_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_dcd_frequency'))

    x_namd_inout_control_velocity_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_dcd_first_step'))

    x_namd_inout_control_force_dcd_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_force_dcd_filename'))

    x_namd_inout_control_force_dcd_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_force_dcd_frequency'))

    x_namd_inout_control_force_dcd_first_step = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_force_dcd_first_step'))

    x_namd_inout_control_output_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_output_filename'))

    x_namd_inout_control_binary_output = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_binary_output'))

    x_namd_inout_control_restart_filename = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_restart_filename'))

    x_namd_inout_control_restart_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_restart_frequency'))

    x_namd_inout_control_binary_restart = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_binary_restart'))

    x_namd_inout_control_switching = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_switching'))

    x_namd_inout_control_switching_on = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_switching_on'))

    x_namd_inout_control_switching_off = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_switching_off'))

    x_namd_inout_control_pairlist_distance = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlist_distance'))

    x_namd_inout_control_pairlist_shrink_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlist_shrink_rate'))

    x_namd_inout_control_pairlist_grow_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlist_grow_rate'))

    x_namd_inout_control_pairlist_trigger = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlist_trigger'))

    x_namd_inout_control_pairlists_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlists_per_cycle'))

    x_namd_inout_control_pairlists = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pairlists'))

    x_namd_inout_control_margin = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_margin'))

    x_namd_inout_control_hydrogen_group_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_hydrogen_group_cutoff'))

    x_namd_inout_control_patch_dimension = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_patch_dimension'))

    x_namd_inout_control_energy_output_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_energy_output_steps'))

    x_namd_inout_control_crossterm_energy = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_crossterm_energy'))

    x_namd_inout_control_timing_output_steps = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_timing_output_steps'))

    x_namd_inout_control_velocity_rescale_freq = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_rescale_freq'))

    x_namd_inout_control_velocity_rescale_temp = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_rescale_temp'))

    x_namd_inout_control_velocity_reassignment_freq = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_reassignment_freq'))

    x_namd_inout_control_velocity_reassignment_temp = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_reassignment_temp'))

    x_namd_inout_control_velocity_reassignment_incr = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_reassignment_incr'))

    x_namd_inout_control_velocity_reassignment_hold = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_reassignment_hold'))

    x_namd_inout_control_lowe_andersen_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_lowe_andersen_dynamics'))

    x_namd_inout_control_lowe_andersen_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_lowe_andersen_temperature'))

    x_namd_inout_control_lowe_andersen_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_lowe_andersen_rate'))

    x_namd_inout_control_lowe_andersen_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_lowe_andersen_cutoff'))

    x_namd_inout_control_langevin_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_dynamics'))

    x_namd_inout_control_langevin_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_temperature'))

    x_namd_inout_control_langevin_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_integrator'))

    x_namd_inout_control_langevin_damping_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_damping_file'))

    x_namd_inout_control_langevin_damping_column = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_damping_column'))

    x_namd_inout_control_langevin_damping_coefficient_unit = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_damping_coefficient_unit'))

    x_namd_inout_control_langevin_dynamics_not_applied_to = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_dynamics_not_applied_to'))

    x_namd_inout_control_temperature_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_temperature_coupling'))

    x_namd_inout_control_coupling_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_coupling_temperature'))

    x_namd_inout_control_berendsen_pressure_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_berendsen_pressure_coupling'))

    x_namd_inout_control_berendsen_compressibility_estimate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_berendsen_compressibility_estimate'))

    x_namd_inout_control_berendsen_relaxation_time = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_berendsen_relaxation_time'))

    x_namd_inout_control_berendsen_coupling_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_berendsen_coupling_frequency'))

    x_namd_inout_control_langevin_piston_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_piston_pressure_control'))

    x_namd_inout_control_target_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_target_pressure'))

    x_namd_inout_control_langevin_oscillation_period = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_oscillation_period'))

    x_namd_inout_control_langevin_decay_time = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_decay_time'))

    x_namd_inout_control_langevin_piston_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_langevin_piston_temperature'))

    x_namd_inout_control_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pressure_control'))

    x_namd_inout_control_initial_strain_rate = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_initial_strain_rate'))

    x_namd_inout_control_cell_fluctuation = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_cell_fluctuation'))

    x_namd_inout_control_particle_mesh_ewald = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_particle_mesh_ewald'))

    x_namd_inout_control_pme_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_tolerance'))

    x_namd_inout_control_pme_ewald_coefficient = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_ewald_coefficient'))

    x_namd_inout_control_pme_interpolation_order = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_interpolation_order'))

    x_namd_inout_control_pme_grid_dimensions = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_grid_dimensions'))

    x_namd_inout_control_pme_maximum_grid_spacing = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_pme_maximum_grid_spacing'))

    x_namd_inout_control_fftw_data_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_fftw_data_file'))

    x_namd_inout_control_full_electrostatic_evaluation_frequency = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_full_electrostatic_evaluation_frequency'))

    x_namd_inout_control_minimization = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_minimization'))

    x_namd_inout_control_velocity_quenching = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_velocity_quenching'))

    x_namd_inout_control_verlet_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_verlet_integrator'))

    x_namd_inout_control_random_number_seed = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_random_number_seed'))

    x_namd_inout_control_use_hydrogen_bonds = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_use_hydrogen_bonds'))

    x_namd_inout_control_coordinate_pdb = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_coordinate_pdb'))

    x_namd_inout_control_structure_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_structure_file'))

    x_namd_inout_control_parameter_file = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_parameter_file'))

    x_namd_inout_control_number_of_parameters = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_number_of_parameters'))

    x_namd_inout_control_parameters = Quantity(
        type=str,
        shape=['x_namd_inout_control_number_of_parameters'],
        description='''
        NAMD running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_inout_control_parameters'))


class x_namd_section_atom_to_atom_type_ref(MSection):
    '''
    Section to store atom label to atom type definition list
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_namd_section_atom_to_atom_type_ref'))

    x_namd_atom_to_atom_type_ref = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_atoms_per_type'],
        description='''
        Reference to the atoms of each atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_to_atom_type_ref'))


class x_namd_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_namd_section_single_configuration_calculation'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_namd_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_positions_image_index'))

    x_namd_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_positions_scaled'))

    x_namd_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_positions_wrapped'))

    x_namd_lattice_lengths = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Lattice dimensions in a vector. Vector includes [a, b, c] lengths.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_namd_lattice_lengths'))

    x_namd_lattice_angles = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Angles of lattice vectors. Vector includes [alpha, beta, gamma] in degrees.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_namd_lattice_angles'))

    x_namd_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''',
        a_legacy=LegacyDefinition(name='x_namd_dummy'))

    x_namd_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''',
        a_legacy=LegacyDefinition(name='x_namd_mdin_finline'))

    x_namd_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_timestep_store'))

    x_namd_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_number_of_atoms_store'))

    x_namd_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_box_bound_store'))

    x_namd_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_box_bounds_store'))

    x_namd_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_variables_store'))

    x_namd_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_traj_atoms_store'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_namd_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_barostat],
        a_legacy=LegacyDefinition(name='x_namd_barostat_target_pressure'))

    x_namd_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_barostat],
        a_legacy=LegacyDefinition(name='x_namd_barostat_tau'))

    x_namd_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_barostat],
        a_legacy=LegacyDefinition(name='x_namd_barostat_type'))

    x_namd_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_namd_integrator_dt'))

    x_namd_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_namd_integrator_type'))

    x_namd_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_namd_periodicity_type'))

    x_namd_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_namd_langevin_gamma'))

    x_namd_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''',
        categories=[public.settings_molecular_dynamics, public.settings_sampling, public.settings_integrator],
        a_legacy=LegacyDefinition(name='x_namd_number_of_steps_requested'))

    x_namd_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_namd_thermostat_level'))

    x_namd_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_namd_thermostat_target_temperature'))

    x_namd_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_namd_thermostat_tau'))

    x_namd_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_namd_thermostat_type'))


class section_atom_type(common.section_atom_type):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_atom_type'))

    x_namd_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Atom name of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_name'))

    x_namd_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_type'))

    x_namd_atom_element = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_element'))

    x_namd_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_type_element'))

    x_namd_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_atom_type_radius'))

    number_of_atoms_per_type = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_type'))


class section_interaction(common.section_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_interaction'))

    x_namd_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_interaction_atom_to_atom_type_ref'))

    x_namd_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_defined_pair_interactions'))

    x_namd_pair_interaction_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_namd_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_pair_interaction_atom_type_ref'))

    x_namd_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_pair_interaction_parameters'))


class section_molecule_interaction(common.section_molecule_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_molecule_interaction'))

    x_namd_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_molecule_interaction_atom_to_atom_type_ref'))

    x_namd_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_defined_molecule_pair_interactions'))

    x_namd_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_pair_molecule_interaction_parameters'))

    x_namd_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_namd_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_pair_molecule_interaction_to_atom_type_ref'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_namd_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_version_date'))

    x_namd_parallel_task_nr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Program task no.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_parallel_task_nr'))

    x_namd_build_osarch = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build OS/ARCH
        ''',
        a_legacy=LegacyDefinition(name='x_namd_build_osarch'))

    x_namd_program_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build date
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_build_date'))

    x_namd_program_citation = Quantity(
        type=str,
        shape=[],
        description='''
        Program citations
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_citation'))

    x_namd_number_of_tasks = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of tasks in parallel program (MPI).
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_tasks'))

    x_namd_program_module_version = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD program module version.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_module_version'))

    x_namd_program_license = Quantity(
        type=str,
        shape=[],
        description='''
        NAMD program license.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_license'))

    x_namd_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''',
        a_legacy=LegacyDefinition(name='x_namd_xlo_xhi'))

    x_namd_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_file_store'))

    x_namd_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_working_path'))

    x_namd_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_execution_host'))

    x_namd_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_execution_path'))

    x_namd_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_module'))

    x_namd_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_execution_date'))

    x_namd_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_program_execution_time'))

    x_namd_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_mdin_header'))

    x_namd_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_mdin_wt'))

    x_namd_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_namd_section_input_output_files'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_namd_section_input_output_files'))

    x_namd_section_control_parameters = SubSection(
        sub_section=SectionProxy('x_namd_section_control_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_namd_section_control_parameters'))


class section_topology(common.section_topology):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_topology'))

    x_namd_input_units_store = Quantity(
        type=str,
        shape=[],
        description='''
        It determines the units of all quantities specified in the input script and data
        file, as well as quantities output to the screen, log file, and dump files.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_input_units_store'))

    x_namd_data_bond_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_bond_types_store'))

    x_namd_data_bond_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_bond_count_store'))

    x_namd_data_angle_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_angle_count_store'))

    x_namd_data_atom_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_atom_types_store'))

    x_namd_data_dihedral_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_dihedral_count_store'))

    x_namd_data_angles_store = Quantity(
        type=str,
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_angles_store'))

    x_namd_data_angle_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_angle_list_store'))

    x_namd_data_bond_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_bond_list_store'))

    x_namd_data_dihedral_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_dihedral_list_store'))

    x_namd_data_dihedral_coeff_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_dihedral_coeff_list_store'))

    x_namd_masses_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_masses_store'))

    x_namd_data_topo_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_namd_data_topo_list_store'))

    x_namd_section_atom_to_atom_type_ref = SubSection(
        sub_section=SectionProxy('x_namd_section_atom_to_atom_type_ref'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_namd_section_atom_to_atom_type_ref'))


class section_frame_sequence(public.section_frame_sequence):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_frame_sequence'))

    x_namd_number_of_volumes_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of volumes in this sequence of frames, see
        x_namd_frame_sequence_volume.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_volumes_in_sequence'))

    x_namd_number_of_densities_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of densities in this sequence of frames, see
        x_namd_frame_sequence_density.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_densities_in_sequence'))

    x_namd_number_of_bond_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_namd_frame_sequence_bond_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_bond_energies_in_sequence'))

    x_namd_number_of_angle_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of angle_energies in this sequence of frames, see
        x_namd_frame_sequence_angle_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_angle_energies_in_sequence'))

    x_namd_number_of_proper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of proper_dihedral_energies in this sequence of frames, see
        x_namd_frame_sequence_proper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_proper_dihedral_energies_in_sequence'))

    x_namd_number_of_improper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of improper_dihedral_energies in this sequence of frames, see
        x_namd_frame_sequence_improper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_improper_dihedral_energies_in_sequence'))

    x_namd_number_of_cmap_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of cmap_dihedral_energies in this sequence of frames, see
        x_namd_frame_sequence_cmap_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_cmap_dihedral_energies_in_sequence'))

    x_namd_number_of_vdw_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of vdw_energies in this sequence of frames, see
        x_namd_frame_sequence_vdw_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_vdw_energies_in_sequence'))

    x_namd_number_of_boundary_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of boundary_energies in this sequence of frames, see
        x_namd_frame_sequence_boundary_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_boundary_energies_in_sequence'))

    x_namd_number_of_electrostatic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of electrostatic_energies in this sequence of frames, see
        x_namd_frame_sequence_electrostatic_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_electrostatic_energies_in_sequence'))

    x_namd_number_of_total2_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total2_energies in this sequence of frames, see
        x_namd_frame_sequence_total2_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_total2_energies_in_sequence'))

    x_namd_number_of_total3_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total3_energies in this sequence of frames, see
        x_namd_frame_sequence_total3_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_total3_energies_in_sequence'))

    x_namd_number_of_misc_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of misc_energies in this sequence of frames, see
        x_namd_frame_sequence_misc_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_number_of_misc_energies_in_sequence'))

    x_namd_frame_sequence_density_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_densities_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_densities values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_density_frames'))

    x_namd_frame_sequence_density = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_densities_in_sequence'],
        description='''
        Array containing the values of the density along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_density_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_density'))

    x_namd_frame_sequence_cmap_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_cmap_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_cmap_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_cmap_dihedral_energy_frames'))

    x_namd_frame_sequence_cmap_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_cmap_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the cmap_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_cmap_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_cmap_dihedral_energy'))

    x_namd_frame_sequence_improper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_improper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_improper_dihedral_energy_frames'))

    x_namd_frame_sequence_improper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the improper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_improper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_improper_dihedral_energy'))

    x_namd_frame_sequence_proper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_proper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_proper_dihedral_energy_frames'))

    x_namd_frame_sequence_proper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the proper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_proper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_proper_dihedral_energy'))

    x_namd_frame_sequence_bond_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_bond_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_bond_energy_frames'))

    x_namd_frame_sequence_bond_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the values of the bond_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_bond_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_bond_energy'))

    x_namd_frame_sequence_boundary_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_boundary values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_boundary_frames'))

    x_namd_frame_sequence_boundary = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the values of the boundary along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_boundary_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_boundary'))

    x_namd_frame_sequence_angle_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_angle_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_angle_energy_frames'))

    x_namd_frame_sequence_angle_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the values of the angle_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_angle_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_angle_energy'))

    x_namd_frame_sequence_vdw_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_vdw_energy values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_vdw_energy_frames'))

    x_namd_frame_sequence_vdw_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the values of the vdw_energy along this sequence of frames (i.e.,
        a trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_vdw_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_vdw_energy'))

    x_namd_frame_sequence_electrostatic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_electrostatic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_electrostatic_energy_frames'))

    x_namd_frame_sequence_electrostatic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the values of the electrostatic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_electrostatic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_electrostatic_energy'))

    x_namd_frame_sequence_total2_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_total2_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_total2_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_total2_energy_frames'))

    x_namd_frame_sequence_total2_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_total2_energies_in_sequence'],
        description='''
        Array containing the values of the total2_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_total2_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_total2_energy'))

    x_namd_frame_sequence_total3_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_total3_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_total3_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_total3_energy_frames'))

    x_namd_frame_sequence_total3_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_total3_energies_in_sequence'],
        description='''
        Array containing the values of the total3_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_total3_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_total3_energy'))

    x_namd_frame_sequence_misc_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_misc_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_misc_energy_frames'))

    x_namd_frame_sequence_misc_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the values of the misc_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_misc_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_misc_energy'))

    x_namd_frame_sequence_volume_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_namd_number_of_volumes_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_namd_frame_sequence_volume values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_volume_frames'))

    x_namd_frame_sequence_volume = Quantity(
        type=np.dtype(np.float64),
        shape=['x_namd_number_of_volumes_in_sequence'],
        description='''
        Array containing the values of the volume along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_volume_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_namd_frame_sequence_volume'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_namd_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_namd_section_single_configuration_calculation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_namd_section_single_configuration_calculation'))


m_package.__init_metainfo__()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fastapi import FastAPI, status, Response
from fastapi.responses import JSONResponse, HTMLResponse
# We use a2wsgi. It is an alternative to the fastapi provided WSGIMiddleware that manages
# to stream requests instead of buffering them.
from a2wsgi import WSGIMiddleware
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.exceptions import HTTPException as StarletteHTTPException
from fastapi.exception_handlers import http_exception_handler as default_http_exception_handler

from nomad import config, infrastructure

from .optimade import optimade_app
from .flask import app as flask_app
from .v1.main import app as v1_app


class OasisAuthenticationMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        path = request.url.path
        if 'extensions' in path or 'info' in path or 'versions' in path:
            return await call_next(request)

        if 'Authorization' not in request.headers:
            return Response(
                status_code=status.HTTP_401_UNAUTHORIZED,
                content='You have to authenticate to use this Oasis endpoint.')

        else:
            user, _ = infrastructure.keycloak.auth(request.headers)
            if user is None or user.email not in config.oasis.allowed_users:
                return Response(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    content='You are not authorized to access this Oasis endpoint.')

        return await call_next(request)


app = FastAPI()

if config.oasis.allowed_users is not None:
    optimade_app.add_middleware(OasisAuthenticationMiddleware)

app_base = config.services.api_base_path
app.mount(f'{app_base}/api/v1', v1_app)
app.mount(f'{app_base}/optimade', optimade_app)
app.mount(app_base, WSGIMiddleware(flask_app))


@app.on_event('startup')
async def startup_event():
    from nomad import infrastructure
    # each subprocess is supposed disconnect connect again: https://jira.mongodb.org/browse/PYTHON-2090
    try:
        from mongoengine import disconnect
        disconnect()
    except Exception:
        pass

    infrastructure.setup()


@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(request, exc):
    if exc.status_code != 404:
        return await default_http_exception_handler(request, exc)

    try:
        accept = request.headers['accept']
    except Exception:
        accept = None

    if accept is not None and 'html' in accept:
        return HTMLResponse(status_code=404, content=f'''
        <html>
            <head><title>{config.meta.name}</title></head>
            <body>
                <h1>NOMAD app</h1>
                <h2>info</h2>
                {'<br/>'.join(f'{key}: {value}' for key, value in config.meta.items())}
                <h2>apis</h2>
                <a href="{app_base}/api">NOMAD API v0</a><br/>
                <a href="{app_base}/api/v1/extensions/docs">NOMAD API v1</a><br/>
                <a href="{app_base}/optimade/v1/extensions/docs">Optimade API</a><br/>
                <a href="{app_base}/dcat">DCAT API</a><br/>
            </body>
        </html>
        ''')

    return JSONResponse(status_code=404, content={
        'detail': 'Not found',
        'info': {
            'app': config.meta,
            'apis': {
                'v0': {
                    'root': f'{app_base}/api',
                    'dashboard': f'{app_base}/api',
                },
                'v1': {
                    'root': f'{app_base}/api/v1',
                    'dashboard': f'{app_base}/api/v1/extensions/docs',
                    'documentation': f'{app_base}/api/v1/extensions/redoc',
                },
                'optimade': {
                    'root': f'{app_base}/optimade/v1',
                    'dashboard': f'{app_base}/optimade/v1/extensions/docs'
                },
                'dcat': {
                    'root': f'{app_base}/dcat',
                    'dashboard': f'{app_base}/dcat'
                }
            }
        }
    })



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The API is protected with *keycloak* and *OpenIDConnect*. All API endpoints that require
or support authentication accept OIDC bearer tokens via HTTP header (``Authentication``).
These token can be acquired from the NOMAD keycloak server or through the ``/auth`` endpoint
that also supports HTTP Basic authentication and passes the given credentials to
keycloak. For GUI's it is recommended to accquire an access token through the regular OIDC
login flow.

Authenticated user information is available via FLASK's build in flask.g.user object.
It is set to None, if no user information is available. To protect endpoints use the following
decorator.

.. autofunction:: authenticate

To allow authentification with signed urls, use this decorator:

.. autofunction:: with_signature_token
'''
from flask import g, request
from flask_restplus import abort, Resource, fields
import functools
import jwt
import datetime
import hmac
import hashlib
import uuid

from nomad import config, processing, utils, infrastructure, datamodel
from nomad.metainfo.flask_extension import generate_flask_restplus_model

from .api import api


# Authentication scheme definitions, for swagger
api.authorizations = {
    'HTTP Basic Authentication': {
        'type': 'basic'
    },
    'OpenIDConnect Bearer Token': {
        'type': 'apiKey',
        'in': 'header',
        'name': 'Authorization'
    },
    'NOMAD upload token': {
        'type': 'apiKey',
        'in': 'query',
        'name': 'token'
    },
    'NOMAD signature': {
        'type': 'apiKey',
        'in': 'query',
        'name': 'signature_token'
    }
}


def authenticate(
        basic: bool = False, upload_token: bool = False, signature_token: bool = False,
        required: bool = False, admin_only: bool = False):
    '''
    A decorator to protect API endpoints with authentication. Uses keycloak access
    token to authenticate users. Other methods might apply. Will abort with 401
    if necessary.

    Arguments:
        basic: Also allow Basic HTTP authentication
        upload_token: Also allow upload_token
        signature_token: Also allow signed urls
        required: Authentication is required
        admin_only: Only the admin user is allowed to use the endpoint.
    '''
    methods = ['OpenIDConnect Bearer Token']
    if basic:
        methods.append('HTTP Basic Authentication')
    if upload_token:
        methods.append('NOMAD upload token')
    if signature_token:
        methods.append('NOMAD signature')

    def decorator(func):
        @functools.wraps(func)
        @api.response(401, 'Not authorized, some data require authentication and authorization')
        @api.doc(security=methods)
        def wrapper(*args, **kwargs):
            g.user = None

            if upload_token and 'token' in request.args:
                try:
                    token = request.args['token']
                    payload, signature = token.split('.')
                    payload = utils.base64_decode(payload)
                    signature = utils.base64_decode(signature)

                    compare = hmac.new(
                        bytes(config.services.api_secret, 'utf-8'),
                        msg=payload,
                        digestmod=hashlib.sha1)

                    if signature != compare.digest():
                        return None

                    user_id = str(uuid.UUID(bytes=payload))
                    g.user = infrastructure.keycloak.get_user(user_id)
                except KeyError:
                    abort(401, 'Invalid token')

            elif signature_token and 'signature_token' in request.args:
                token = request.args.get('signature_token', None)
                try:
                    decoded = jwt.decode(token, config.services.api_secret, algorithms=['HS256'])
                    g.user = datamodel.User.get(user_id=decoded['user'])
                except KeyError:
                    abort(401, 'Token with invalid/unexpected payload')
                except jwt.ExpiredSignatureError:
                    abort(401, 'Expired token')
                except jwt.InvalidTokenError:
                    abort(401, 'Invalid token')

            elif 'token' in request.args:
                abort(401, 'Query param token not supported for this endpoint')

            elif 'signature_token' in request.args:
                abort(401, 'Query param signature_token not supported for this endpoint')

            else:
                try:
                    g.user, g.oidc_access_token = infrastructure.keycloak.auth(request.headers, allow_basic=basic)
                except infrastructure.KeycloakError as e:
                    abort(401, message=str(e))

            if config.oasis.allowed_users is not None:
                if g.user is None:
                    abort(401, message='Authentication is required for this Oasis')
                if g.user.email not in config.oasis.allowed_users:
                    abort(401, message='You are not authorized to access this Oasis')

            if required and g.user is None:
                abort(401, message='Authentication is required for this endpoint')
            if admin_only and (g.user is None or not g.user.is_admin):
                abort(401, message='Only the admin user is allowed to use this endpoint')

            return func(*args, **kwargs)

        return wrapper

    return decorator


def generate_upload_token(user):
    payload = uuid.UUID(user.user_id).bytes
    signature = hmac.new(
        bytes(config.services.api_secret, 'utf-8'),
        msg=payload,
        digestmod=hashlib.sha1)

    return '%s.%s' % (
        utils.base64_encode(payload),
        utils.base64_encode(signature.digest()))


ns = api.namespace(
    'auth',
    description='Authentication related endpoints.')


auth_model = api.model('Auth', {
    'access_token': fields.String(description='The OIDC access token'),
    'upload_token': fields.String(description='A short token for human readable upload URLs'),
    'signature_token': fields.String(description='A short term token to sign URLs')
})


@ns.route('/')
class AuthResource(Resource):
    @api.doc('get_auth')
    @api.marshal_with(auth_model, skip_none=True, code=200, description='Auth info send')
    @authenticate(required=True, basic=True)
    def get(self):
        '''
        Provides authentication information. This endpoint requires authentification.
        Like all endpoints the OIDC access token based authentification. In additional,
        basic HTTP authentification can be used. This allows to login and acquire an
        access token.

        The response contains a short (10s) term JWT token that can be used to sign
        URLs with a ``signature_token`` query parameter, e.g. for file downloads on the
        raw or archive api endpoints; a short ``upload_token`` that is used in
        ``curl`` command line based uploads; and the OIDC JWT access token.
        '''

        def signature_token():
            expires_at = datetime.datetime.utcnow() + datetime.timedelta(seconds=10)
            return jwt.encode(
                dict(user=g.user.user_id, exp=expires_at),
                config.services.api_secret, 'HS256').decode('utf-8')

        try:
            return {
                'upload_token': generate_upload_token(g.user),
                'signature_token': signature_token(),
                'access_token': g.oidc_access_token
            }

        except KeyError:
            abort(401, 'The authenticated user does not exist')


user_model = generate_flask_restplus_model(api, datamodel.User.m_def)
users_model = api.model('UsersModel', {
    'users': fields.Nested(user_model, skip_none=True)
})


users_parser = api.parser()
users_parser.add_argument(
    'query', default='',
    help='Only return users that contain this string in their names, usernames, or emails.')


@ns.route('/users')
class UsersResource(Resource):
    @api.doc('get_users')
    @api.marshal_with(users_model, code=200, description='User suggestions send')
    @api.expect(users_parser, validate=True)
    def get(self):
        ''' Get existing users. '''
        args = users_parser.parse_args()

        return dict(users=infrastructure.keycloak.search_user(args.get('query')))

    @api.doc('invite_user')
    @api.marshal_with(user_model, code=200, skip_none=True, description='User invited')
    @api.expect(user_model, validate=True)
    def put(self):
        ''' Invite a new user. '''
        if config.keycloak.oasis:
            abort(400, 'User invide does not work this NOMAD OASIS')

        json_data = request.get_json()
        try:
            user = datamodel.User.m_from_dict(json_data)
        except Exception as e:
            abort(400, 'Invalid user data: %s' % str(e))

        if user.email is None:
            abort(400, 'Invalid user data: email is required')

        try:
            error = infrastructure.keycloak.add_user(user, invite=True)
        except KeyError as e:
            abort(400, 'Invalid user data: %s' % str(e))

        if error is not None:
            abort(400, 'Could not invite user: %s' % error)

        return datamodel.User.get(username=user.username), 200


def with_signature_token(func):
    '''
    A decorator for API endpoint implementations that validates signed URLs. Token to
    sign URLs can be retrieved via the ``/auth`` endpoint.
    '''
    @functools.wraps(func)
    @api.response(401, 'Invalid or expired signature token')
    def wrapper(*args, **kwargs):
        token = request.args.get('token', None)
        if token is not None:
            try:
                decoded = jwt.decode(token, config.services.api_secret, algorithms=['HS256'])
                user = datamodel.User.get(decoded['user'])
                if user is None:
                    abort(401, 'User for token does not exist')
                else:
                    g.user = user
            except KeyError:
                abort(401, 'Token with invalid/unexpected payload')
            except jwt.ExpiredSignatureError:
                abort(401, 'Expired token')
            except jwt.InvalidTokenError:
                abort(401, 'Invalid token')

        return func(*args, **kwargs)

    return wrapper


def create_authorization_predicate(upload_id, calc_id=None):
    '''
    Returns a predicate that determines if the logged in user has the authorization
    to access the given upload and calculation.
    '''
    def func():
        if g.user is None:
            # guest users don't have authorized access to anything
            return False
        elif g.user.is_admin:
            # the admin user does have authorization to access everything
            return True

        # look in mongo
        try:
            upload = processing.Upload.get(upload_id)
            if g.user.user_id == upload.user_id:
                return True

            if calc_id is not None:
                try:
                    calc = processing.Calc.get(calc_id)
                except KeyError:
                    return False
                return g.user.user_id in calc.metadata.get('shared_with', [])

            return False

        except KeyError as e:
            logger = utils.get_logger(__name__, upload_id=upload_id, calc_id=calc_id)
            logger.error('Upload files without respective db entry')
            raise e

    return func



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import json
import logging

from nomad.utils import configure_logging
from nomad.datamodel import EntryArchive
from bandparser import BandParser


if __name__ == "__main__":
    configure_logging(console_log_level=logging.DEBUG)
    archive = EntryArchive()
    BandParser().parse(sys.argv[1], archive, logging)
    json.dump(archive.m_to_dict(), sys.stdout, indent=2)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import io
from datetime import datetime
from typing import List, Dict, Any, Optional
from pydantic import BaseModel, Field, validator
from fastapi import (
    APIRouter, Request, File, UploadFile, status, Depends, Path, Query as FastApiQuery,
    HTTPException)

from nomad import utils, config, files, datamodel
from nomad.processing import Upload, Calc, ProcessAlreadyRunning, FAILURE
from nomad.processing.base import PROCESS_COMPLETED
from nomad.utils import strip

from .auth import create_user_dependency, generate_upload_token
from ..models import (
    BaseModel, User, Direction, Pagination, PaginationResponse, HTTPExceptionModel)
from ..utils import parameter_dependency_from_model, create_responses

router = APIRouter()
default_tag = 'uploads'

logger = utils.get_logger(__name__)


_not_authorized = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. Authorization is required, but no or bad authentication credentials provided.''')}

_not_authorized_to_upload = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. No credentials provided, or you do not have permissions to the
        specified upload.''')}

_not_authorized_to_entry = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. No credentials provided, or you do not have permissions to the
        specified upload or entry.''')}

_bad_request = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Bad request. The request could not be processed because of some error/invalid argument.''')}

_bad_pagination = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Bad request. Invalid pagination arguments supplied.''')}

_upload_not_found = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The specified upload could not be found.''')}

_entry_not_found = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The specified upload or entry could not be found.''')}


class ProcData(BaseModel):
    tasks: List[str] = Field()
    tasks_running: bool = Field()
    tasks_status: str = Field()
    current_task: Optional[str] = Field()
    process_running: bool = Field()
    current_process: Optional[str] = Field()
    errors: List[str] = Field()
    warnings: List[str] = Field()
    create_time: datetime = Field()
    complete_time: Optional[datetime] = Field()

    class Config:
        orm_mode = True


class UploadProcData(ProcData):
    upload_id: str = Field(
        None,
        description='The unique id for the upload.')
    name: Optional[str] = Field(
        description='The name of the upload. This can be provided during upload '
                    'using the name query parameter.')
    upload_time: datetime = Field(
        None,
        description='The time of upload.')
    upload_path: Optional[str] = Field(
        description='Path to the uploaded file on the server.')
    published: bool = Field(
        False,
        description='If this upload is already published.')
    published_to: List[str] = Field(
        None,
        description='A list of other NOMAD deployments that this upload was uploaded to already.')
    last_status_message: Optional[str] = Field(
        None,
        description='The last informative message that the processing saved about this uploads status.')


class EntryProcData(ProcData):
    entry_id: str = Field()
    mainfile: str = Field()
    upload_id: str = Field()
    parser: str = Field()


class UploadProcDataPagination(Pagination):
    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        if order_by is None:
            return 'create_time'  # Default value
        assert order_by in ('create_time', 'published'), 'order_by must be a valid attribute'
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # Validation handled elsewhere
        return page_after_value


upload_proc_data_pagination_parameters = parameter_dependency_from_model(
    'upload_proc_data_pagination_parameters', UploadProcDataPagination)


class EntryProcDataPagination(Pagination):
    @validator('order_by')
    def validate_order_by(cls, order_by):  # pylint: disable=no-self-argument
        if order_by is None:
            return 'mainfile'  # Default value
        assert order_by in ('mainfile', 'parser', 'tasks_status', 'current_task'), 'order_by must be a valid attribute'
        return order_by

    @validator('page_after_value')
    def validate_page_after_value(cls, page_after_value, values):  # pylint: disable=no-self-argument
        # Validation handled elsewhere
        return page_after_value


entry_proc_data_pagination_parameters = parameter_dependency_from_model(
    'entry_proc_data_pagination_parameters', EntryProcDataPagination)


class UploadProcDataResponse(BaseModel):
    upload_id: str = Field(None, description=strip('''
        Unique id of the upload.'''))
    data: UploadProcData = Field(
        None, description=strip('''
        The upload data as a dictionary.'''))


class UploadProcDataQuery(BaseModel):
    upload_id: Optional[List[str]] = Field(
        description='Search for uploads matching the given id. Multiple values can be specified.')
    upload_name: Optional[List[str]] = Field(
        description='Search for uploads matching the given name. Multiple values can be specified.')
    is_processing: Optional[bool] = Field(
        description=strip('''
            If True, only include currently processing uploads.
            If False, do not include currently processing uploads.
            If unset, include everything.'''))
    is_published: Optional[bool] = Field(
        description=strip('''
            If True: only include published uploads.
            If False: only include unpublished uploads.
            If unset: include everything.'''))


upload_proc_data_query_parameters = parameter_dependency_from_model(
    'upload_proc_data_query_parameters', UploadProcDataQuery)


class UploadProcDataQueryResponse(BaseModel):
    query: UploadProcDataQuery = Field()
    pagination: PaginationResponse = Field()
    data: List[UploadProcData] = Field(
        None, description=strip('''
        The upload data as a list. Each item is a dictionary with the data for each
        upload.'''))


class EntryProcDataResponse(BaseModel):
    entry_id: str = Field()
    data: EntryProcData = Field()


class EntryProcDataQueryResponse(BaseModel):
    pagination: PaginationResponse = Field()
    processing_successful: int = Field(
        None, description=strip('''
        Number of entries that has been processed successfully.
        '''))
    processing_failed: int = Field(
        None, description=strip('''
        Number of entries that failed to process.
        '''))
    data: List[EntryProcData] = Field(
        None, description=strip('''
        The entries data as a list. Each item is a dictionary with the data for one entry.
        '''))


class UploadCommandExamplesResponse(BaseModel):
    upload_url: str = Field()
    upload_command: str = Field()
    upload_command_with_name: str = Field()
    upload_progress_command: str = Field()
    upload_command_form: str = Field()
    upload_tar_command: str = Field()


@router.get(
    '/command-examples', tags=[default_tag],
    summary='Get example commands for shell based uploads.',
    response_model=UploadCommandExamplesResponse,
    responses=create_responses(_not_authorized),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_command_examples(user: User = Depends(create_user_dependency(required=True))):
    ''' Get url and example command for shell based uploads. '''
    token = generate_upload_token(user)
    api_url = config.api_url(ssl=config.services.https_upload, api='api/v1')
    upload_url = f'{api_url}/uploads?token={token}'
    upload_url_with_name = upload_url + '&name=<name>'
    # Upload via streaming data tends to work much easier, e.g. no mime type issues, etc.
    # It is also easier for the user to unterstand IMHO.
    upload_command = f"curl -X POST '{upload_url}' -T <local_file>"
    rv = UploadCommandExamplesResponse(
        upload_url=upload_url,
        upload_command=upload_command,
        upload_command_form=f"curl -X POST '{upload_url}' -F file=@<local_file>",
        upload_command_with_name=f"curl -X POST '{upload_url_with_name}' -T <local_file>",
        upload_progress_command=upload_command + ' | xargs echo',
        upload_tar_command=f"tar -cf - <local_folder> | curl -# '{upload_url}' -X POST -T - | xargs echo"
    )
    return rv


@router.get(
    '', tags=[default_tag],
    summary='List uploads of authenticated user.',
    response_model=UploadProcDataQueryResponse,
    responses=create_responses(_not_authorized, _bad_pagination),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_uploads(
        request: Request,
        query: UploadProcDataQuery = Depends(upload_proc_data_query_parameters),
        pagination: UploadProcDataPagination = Depends(upload_proc_data_pagination_parameters),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Retrieves metadata about all uploads that match the given query criteria.
    '''
    # Build query
    query_kwargs: Dict[str, Any] = {}
    query_kwargs.update(user_id=str(user.user_id))

    if query.upload_id:
        query_kwargs.update(upload_id__in=query.upload_id)

    if query.upload_name:
        query_kwargs.update(name__in=query.upload_name)

    if query.is_processing is True:
        query_kwargs.update(process_status__ne=PROCESS_COMPLETED)
    elif query.is_processing is False:
        query_kwargs.update(process_status=PROCESS_COMPLETED)

    if query.is_published is True:
        query_kwargs.update(published=True)
    elif query.is_published is False:
        query_kwargs.update(published=False)

    # Fetch data from DB
    mongodb_query = _query_mongodb(**query_kwargs)
    # Create response
    start = pagination.get_simple_index()
    end = start + pagination.page_size

    order_by = pagination.order_by
    order_by_with_sign = order_by if pagination.order == Direction.asc else '-' + order_by
    if order_by == 'create_time':
        order_by_args = [order_by_with_sign, 'upload_id']  # Use upload_id as tie breaker
    elif order_by == 'published':
        order_by_args = [order_by_with_sign, 'create_time', 'upload_id']

    mongodb_query = mongodb_query.order_by(*order_by_args)

    data = [_upload_to_pydantic(upload) for upload in mongodb_query[start:end]]

    pagination_response = PaginationResponse(total=mongodb_query.count(), **pagination.dict())
    pagination_response.populate_simple_index_and_urls(request)

    return UploadProcDataQueryResponse(
        query=query,
        pagination=pagination_response,
        data=data)


@router.get(
    '/{upload_id}', tags=[default_tag],
    summary='Get a specific upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to retrieve.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Fetches a specific upload by its upload_id.
    '''
    # Get upload (or throw exception if nonexistent/no access)
    upload = _get_upload_with_read_access(upload_id, user)

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.get(
    '/{upload_id}/entries', tags=[default_tag],
    summary='Get the entries of the specific upload as a list',
    response_model=EntryProcDataQueryResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_pagination),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_entries(
        request: Request,
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to retrieve entries for.'),
        pagination: EntryProcDataPagination = Depends(entry_proc_data_pagination_parameters),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Fetches the entries of a specific upload. Pagination is used to browse through the
    results.
    '''
    upload = _get_upload_with_read_access(upload_id, user)

    order_by = pagination.order_by
    order_by_with_sign = order_by if pagination.order == Direction.asc else '-' + order_by

    start = pagination.get_simple_index()
    end = start + pagination.page_size

    # load upload's calcs. Use calc_id as tie breaker for ordering.
    entries = list(upload.all_calcs(start, end, order_by=(order_by_with_sign, 'calc_id')))
    failed_calcs = upload.failed_calcs

    pagination_response = PaginationResponse(total=upload.total_calcs, **pagination.dict())
    pagination_response.populate_simple_index_and_urls(request)
    return EntryProcDataQueryResponse(
        pagination=pagination_response,
        processing_successful=upload.processed_calcs - failed_calcs,
        processing_failed=failed_calcs,
        data=[_entry_to_pydantic(entry) for entry in entries])


@router.get(
    '/{upload_id}/entries/{entry_id}', tags=[default_tag],
    summary='Get a specific entry for a specific upload',
    response_model=EntryProcDataResponse,
    responses=create_responses(_entry_not_found, _not_authorized_to_entry),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_upload_entry(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload.'),
        entry_id: str = Path(
            ...,
            description='The unique id of the entry, belonging to the specified upload.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Fetches a specific entry for a specific upload.
    '''
    upload = _get_upload_with_read_access(upload_id, user)
    entry = upload.get_calc(entry_id)
    if not entry:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            An entry by that id could not be found in the specified upload.'''))
    return EntryProcDataResponse(
        entry_id=entry_id,
        data=_entry_to_pydantic(entry))


@router.post(
    '', tags=[default_tag],
    summary='Submit a new upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_not_authorized, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload(
        request: Request,
        file: UploadFile = File(None),
        local_path: str = None,  # Internal use/admins only
        name: str = FastApiQuery(
            None,
            description=strip('''
            Specifies the name of the upload.''')),
        publish_directly: bool = FastApiQuery(
            None,
            description=strip('''
            If the upload should be published directly. False by default.''')),
        oasis_upload_id: str = FastApiQuery(
            None,
            description=strip('''
            For oasis uploads: the upload id of the oasis system.''')),
        oasis_uploader_id: str = FastApiQuery(
            None,
            description=strip('''
            For oasis uploads: the id of the user in the oasis system who made the upload
            originally. The user must also be registered in NOMAD.''')),
        oasis_deployment_id: str = FastApiQuery(
            None,
            description=strip('''
            For oasis uploads: the deployment id.''')),
        user: User = Depends(create_user_dependency(required=True, upload_token_auth_allowed=True))):
    '''
    Upload a file to the repository. Can be used to upload files via browser or other
    http clients like curl. This will also start the processing of the upload.

    There are two basic ways to upload a file: in the multipart-formdata or streaming the
    file data in the http body. Both are supported. The second method does not transfer a
    filename, so it is recommended to supply the parameter `name` in this case.

    Method 1: multipart-formdata

        curl -X 'POST' "url" -F file=@local_file

    Method 2: streaming data

        curl -X 'POST' "url" -T local_file

    Authentication is required to perform an upload. This can either be done using the
    regular bearer token, or using the simplified upload token. To use the simplified
    upload token, just specify it as a query parameter in the url, i.e.

        curl -X 'POST' ".../uploads?token=ABC.XYZ" ...

    Note, there is a limit on how many unpublished uploads a user can have. If exceeded,
    error code 400 will be returned.
    '''
    # Determine the source data stream
    src_stream = None
    if local_path:
        # Method 0: Local file - only for admins
        if not user.is_admin:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
                You need to be admin to use local_path as method of upload.'''))
        if not os.path.exists(local_path) or not os.path.isfile(local_path):
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                The specified local_path cannot be found or is not a file.'''))
        if not name:
            name = os.path.basename(local_path)
        src_stream = _asyncronous_file_reader(open(local_path, 'rb'))
    elif file:
        # Method 1: Data provided as formdata
        if not name:
            name = file.filename
        src_stream = _asyncronous_file_reader(file)
    else:
        # Method 2: Data has to be sent streamed in the body
        src_stream = request.stream()

    if not user.is_admin:
        # Check upload limit
        if _query_mongodb(user_id=str(user.user_id), published=False).count() >= config.services.upload_limit:  # type: ignore
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                Limit of unpublished uploads exceeded for user.'''))

    # check if allowed to perform oasis upload
    from_oasis = oasis_upload_id is not None
    if from_oasis:
        if not user.is_oasis_admin:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='Only an oasis admin can perform an oasis upload.')
        if oasis_uploader_id is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='You must provide the original uploader for an oasis upload.')
        if oasis_deployment_id is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='You must provide the oasis deployment id for an oasis upload.')
        if publish_directly is False:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='Oasis uploads are always published directly.')
        # Switch user!
        user = datamodel.User.get(user_id=oasis_uploader_id)
        if user is None:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='The given original uploader does not exist.')
    elif oasis_uploader_id is not None or oasis_deployment_id is not None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='For an oasis upload you must provide an oasis_upload_id.')

    # Get upload_id and path
    if from_oasis:
        upload_id = oasis_upload_id
        try:
            Upload.get(upload_id)
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='An oasis upload with the given upload_id already exists.')
        except KeyError:
            pass
    else:
        upload_id = utils.create_uuid()

    logger.info('upload created', upload_id=upload_id)

    # Read the stream and save to file
    if local_path:
        upload_path = local_path  # Use the provided path
        uploaded_bytes = os.path.getsize(local_path)
    else:
        upload_path = files.PathObject(config.fs.tmp, upload_id).os_path
        try:
            with open(upload_path, 'wb') as f:
                uploaded_bytes = 0
                log_interval = 1e9
                log_unit = 'GB'
                next_log_at = log_interval
                async for chunk in src_stream:
                    if not chunk:
                        # End of data stream
                        break
                    uploaded_bytes += len(chunk)
                    f.write(chunk)
                    if uploaded_bytes > next_log_at:
                        logger.info('Large upload in progress - uploaded: '
                                    f'{uploaded_bytes // log_interval} {log_unit}')
                        next_log_at += log_interval
                logger.info(f'Uploaded {uploaded_bytes} bytes')
        except Exception:
            if os.path.isfile(upload_path):
                os.remove(upload_path)
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                    Some IO went wrong, download probably aborted/disrupted.'''))

    if not uploaded_bytes:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
                Empty upload - not allowed.'''))

    logger.info('received uploaded file')

    upload = Upload.create(
        upload_id=upload_id,
        user=user,
        name=name,
        upload_time=datetime.utcnow(),
        upload_path=upload_path,
        temporary=local_path != upload_path,
        publish_directly=publish_directly or from_oasis,
        from_oasis=from_oasis,
        oasis_deployment_id=oasis_deployment_id)

    upload.process_upload()

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.delete(
    '/{upload_id}', tags=[default_tag],
    summary='Delete an upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def delete_upload(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to delete.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Delete an existing upload.

    Only uploads that are sill in staging, not already deleted, not still uploaded, and
    not currently processed, can be deleted.
    '''
    upload = _get_upload_with_write_access(upload_id, user)

    if upload.tasks_running:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            The upload is not processed yet.'''))

    try:
        upload.delete_upload()
    except ProcessAlreadyRunning:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=strip('''
            The upload is still being processed.'''))
    except Exception as e:
        logger.error('could not delete processing upload', exc_info=e)
        raise

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/action/publish', tags=[default_tag],
    summary='Publish an upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_action_publish(
        upload_id: str = Path(
            ...,
            description=strip('''
                The unique id of the upload to publish.''')),
        with_embargo: bool = FastApiQuery(
            True,
            description=strip('''
                If the data is published with an embargo.''')),
        embargo_length: int = FastApiQuery(
            36,
            description=strip('''
                Length of the requested embargo in months.''')),
        to_central_nomad: bool = FastApiQuery(
            False,
            description=strip('''
                Will send the upload to the central NOMAD repository and publish it. This
                option is only available on an OASIS. The upload must already be published
                on the OASIS.''')),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Publishes an upload. The upload cannot be modified after this point, and after the
    embargo period (if any) is expired, the generated archive entries will be publicly visible.
    '''
    upload = _get_upload_with_write_access(upload_id, user)

    if upload.tasks_running or upload.process_running:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The upload is not finished processing yet.')
    if upload.tasks_status == FAILURE:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Cannot publish an upload that failed processing.')
    if upload.processed_calcs == 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Cannot publish an upload without any resulting entries.')

    if to_central_nomad:
        # Publish from an OASIS to the central repository
        if not config.keycloak.oasis:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='Must be on an OASIS to publish to the central NOMAD repository.')
        if not upload.published:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='The upload must be published on the OASIS first.')
        # Everything looks ok, try to publish it to the central NOMAD!
        upload.publish_from_oasis()
    else:
        # Publish to this repository
        if upload.published:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail='The upload is already published.')
        metadata_dict: Dict[str, Any] = {'with_embargo': with_embargo}
        if with_embargo:
            if not embargo_length or not 0 < embargo_length <= 36:
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail='embargo_length needs to be between 1 and 36 months.')
            metadata_dict.update(embargo_length=embargo_length)
        try:
            upload.compress_and_set_metadata(metadata_dict)
            upload.publish_upload()
        except ProcessAlreadyRunning:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail='The upload is still/already processed.')

    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


@router.post(
    '/{upload_id}/action/re-process', tags=[default_tag],
    summary='Re-process a published upload',
    response_model=UploadProcDataResponse,
    responses=create_responses(_upload_not_found, _not_authorized_to_upload, _bad_request),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_upload_action_reprocess(
        upload_id: str = Path(
            ...,
            description='The unique id of the upload to re-process.'),
        user: User = Depends(create_user_dependency(required=True))):
    '''
    Re-processes an upload. The upload must be published, have at least one outdated
    caclulation, and not be processing at the moment.
    '''
    upload = _get_upload_with_write_access(upload_id, user, published_requires_admin=False)

    if upload.tasks_running or upload.process_running:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The upload is currently being processed.')
    if not upload.published:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Only published uploads can be re-processed.')
    if len(upload.outdated_calcs) == 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='You can only re-process uploads with at least one outdated calculation')

    upload.reset()
    upload.re_process_upload()
    return UploadProcDataResponse(
        upload_id=upload_id,
        data=_upload_to_pydantic(upload))


async def _asyncronous_file_reader(f):
    ''' Asynchronous generator to read file-like objects. '''
    while True:
        try:
            data: bytes = await f.read(io.DEFAULT_BUFFER_SIZE)
        except Exception:
            await f.close()
            raise
        if not data:
            await f.close()
            return
        yield data


def _query_mongodb(**kwargs):
    return Upload.objects(**kwargs)


def _get_upload_with_read_access(upload_id, user) -> Upload:
    '''
    Determines if the specified user has read access to the specified upload. If so, the
    corresponding Upload object is returned. If the upload does not exist, or the user has
    no read access to it, a HTTPException is raised.
    '''
    if not user:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            User authentication required to access uploads.'''))
    mongodb_query = _query_mongodb(upload_id=upload_id)
    if not mongodb_query.count():
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            The specified upload_id was not found.'''))
    upload = mongodb_query.first()
    if user.is_admin or upload.user_id == str(user.user_id):
        # Ok, it exists and belongs to user
        return upload
    else:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            You do not have access to the specified upload.'''))


def _get_upload_with_write_access(upload_id, user, published_requires_admin=True) -> Upload:
    if not user:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            User authentication required to access uploads.'''))
    mongodb_query = _query_mongodb(upload_id=upload_id)
    if not mongodb_query.count():
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=strip('''
            The specified upload_id was not found.'''))
    upload = mongodb_query.first()
    if upload.user_id != str(user.user_id) and not user.is_admin:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            You do not have write access to the specified upload.'''))
    if published_requires_admin and upload.published and not user.is_admin:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            Only admins can change uploads that are published.'''))
    return upload


def _upload_to_pydantic(upload: Upload) -> UploadProcData:
    ''' Converts the mongo db object to an UploadProcData object. '''
    return UploadProcData.from_orm(upload)


def _entry_to_pydantic(entry: Calc) -> EntryProcData:
    ''' Converts the mongo db object to an EntryProcData object'''
    return EntryProcData.from_orm(entry)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os.path
import json
import numpy as np

from nomad.normalizing.normalizer import Normalizer

controlIn_basis_set = 'x_fhi_aims_section_controlIn_basis_set'
controlIn_basis_func = 'x_fhi_aims_section_controlIn_basis_func'
controlIn_nucleus = 'x_fhi_aims_controlIn_nucleus'

pure_types_json = dict()

for pure_types_str in ['light', 'really_tight', 'tight']:
    with open(os.path.join(os.path.dirname(__file__), 'data', pure_types_str + '.json')) as f:
        json_data = json.load(f)
        section_method = json_data['sections']['section_run-0']['sections']['section_method-0']
        pure_types_json[pure_types_str] = section_method[controlIn_basis_set]


class FhiAimsBaseNormalizer(Normalizer):

    # Finds out if val is in the array
    def compare_val_list(self, val, list):
        if val in list:
            return 0
        return 1

    # Comparison between two dicts.
    # List structure:
    def compare_dict_dict(self, d1, d2):
        sum2 = np.zeros(len(d2))

        # Loop over the size of dict2
        for k in np.arange(0, len(d2)):
            # Lopp over the elements of each dict2
            for idx, val in d1.items():
                # Excludes the keys that are always different.
                if (idx not in ["gIndex", "references", "uri"]):
                    try:
                        if (val != d2[k][idx]):
                            sum2[k] = sum2[k] + 1
                    except KeyError:  # this exception case arises if the cut off potential is not a number
                        continue
        if (min(sum2) == 0):
            return 0
        else:
            return 1  # sum(sum2)

    def compare_to_defaults(self, dict2_default, dict1):
        # first compare the integration grid
        false_hits_integration_grid = 0
        false_hits_basis = 0

        for key in dict1:
            if key not in ['gIndex', 'uri', controlIn_basis_func]:
                if np.size(dict1[key]) == 1:
                    if(dict1[key] != dict2_default[key]):
                        false_hits_integration_grid += 1
                        false_hits_integration_grid += abs(np.size(dict1[key]) - np.size(dict2_default[key]))
                if np.size(dict1[key]) > 1:
                    for i in dict1[key]:
                        false_hits_integration_grid += self.compare_val_list(i, dict2_default[key])
                        false_hits_integration_grid += abs(np.size(dict1[key]) - np.size(dict2_default[key]))

            elif (key == controlIn_basis_func):
                for i in np.arange(0, len(dict1[key])):
                    false_hits_basis += self.compare_dict_dict(
                        dict1[key][i], dict2_default[key])
                false_hits_basis += abs(len(dict1[key]) - len(dict2_default[key]))

        return [false_hits_integration_grid, false_hits_basis]

    def normalize(self, logger=None) -> None:
        super().normalize(logger)
        if not self.section_run or self.section_run.program_name != 'FHI-aims':
            return

        for method in self.section_run.section_method:
            to_compare = getattr(method, controlIn_basis_set, None)
            if to_compare is None:
                # not fhi aims data
                continue

            matrix_hits_int = dict.fromkeys(pure_types_json, 0)
            matrix_hits_basis = dict.fromkeys(pure_types_json, 0)

            for index, data in enumerate(to_compare):
                atom_index = int(data[controlIn_nucleus])
                for key, val in pure_types_json.items():
                    results = self.compare_to_defaults(val[atom_index], to_compare[index])
                    matrix_hits_int[key] += results[0]
                    matrix_hits_basis[key] += results[1]

                    # matrix_hits[key]=matrix_hits[key]+CompareToDefaults(val[AtomIndex],to_compare[i])

            closest_base_int = min(matrix_hits_int, key=matrix_hits_int.get)
            if (matrix_hits_basis[min(matrix_hits_basis, key=matrix_hits_basis.get)] == 0):
                closest_base_base = ''
            else:
                closest_base_base = '+'

            if (matrix_hits_int[closest_base_int] == 0):
                method.basis_set = closest_base_int + closest_base_base
            elif(matrix_hits_int[closest_base_int] <= 5):
                method.basis_set = '~' + closest_base_int + closest_base_base
            elif(matrix_hits_int[closest_base_int] > 5):
                method.basis_set = 'custom-' + closest_base_int



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
API endpoint that deliver backend configuration details.
'''

from typing import Dict, Any
from flask_restplus import Resource, fields
from datetime import datetime

from nomad import config, normalizing, datamodel, gitinfo, search
from nomad.parsing import parsers, MatchingParser

from .api import api


ns = api.namespace('info', description='Access to nomad configuration details.')

metainfo_model = api.model('Metainfo', {
    'all_package': fields.String(description='Name of the metainfo package that references all available packages, i.e. the complete metainfo.'),
    'root_section': fields.String(description='Name of the topmost section, e.g. section_run for computational material science data.')
})

domain_model = api.model('Domain', {
    'name': fields.String,
    'metainfo': fields.Nested(model=metainfo_model)
})

git_info_model = api.model('GitInfo', {
    'ref': fields.String,
    'version': fields.String,
    'commit': fields.String,
    'log': fields.String
})

statistics_info_model = api.model('StatisticsInfo', {
    'n_entries': fields.Integer(description='Number of entries in NOMAD'),
    'n_uploads': fields.Integer(description='Number of uploads in NOMAD'),
    'n_quantities': fields.Integer(description='Accumulated number of quantities over all entries in the Archive'),
    'n_calculations': fields.Integer(description='Accumulated number of calculations, e.g. total energy calculations in the Archive'),
    'n_materials': fields.Integer(description='Number of materials in NOMAD')
    # TODO
    # 'raw_file_size': fields.Integer(description='Total amount of raw files in TB'),
    # 'archive_file_size': fields.Integer(description='Total amount of binary archive data in TB')
})

code_info_model = api.model('CodeInfo', {
    'code_name': fields.String(description='Name of the code or input format', allow_null=True),
    'code_homepage': fields.String(description='Homepage of the code or input format', allow_null=True)
}, allow_null=True, skip_none=True)

info_model = api.model('Info', {
    'parsers': fields.List(fields.String),
    'metainfo_packages': fields.List(fields.String),
    'codes': fields.List(fields.Nested(code_info_model)),
    'normalizers': fields.List(fields.String),
    'domains': fields.List(fields.Nested(model=domain_model)),
    'statistics': fields.Nested(model=statistics_info_model, description='General NOMAD statistics'),
    'search_quantities': fields.Raw(),
    'version': fields.String,
    'release': fields.String,
    'git': fields.Nested(model=git_info_model),
    'oasis': fields.Boolean
})


_statistics: Dict[str, Any] = None


def statistics():
    global _statistics
    if _statistics is None or datetime.now().timestamp() - _statistics.get('timestamp', 0) > 3600 * 24:
        _statistics = dict(timestamp=datetime.now().timestamp())
        _statistics.update(
            **search.SearchRequest().global_statistics().execute()['global_statistics'])

    return _statistics


@ns.route('/')
class InfoResource(Resource):
    @api.doc('get_info')
    @api.marshal_with(info_model, skip_none=True, code=200, description='Info send')
    def get(self):
        ''' Return information about the nomad backend and its configuration. '''
        codes_dict = {}
        for parser in parsers.parser_dict.values():
            if isinstance(parser, MatchingParser) and parser.domain == 'dft':
                code_name = parser.code_name
                if code_name in codes_dict:
                    continue
                codes_dict[code_name] = dict(code_name=code_name, code_homepage=parser.code_homepage)
        codes = sorted(list(codes_dict.values()), key=lambda code_info: code_info['code_name'].lower())

        return {
            'parsers': [
                key[key.index('/') + 1:]
                for key in parsers.parser_dict.keys()],
            'metainfo_packages': ['general', 'general.experimental', 'common', 'public'] + sorted([
                key[key.index('/') + 1:]
                for key in parsers.parser_dict.keys()]),
            'codes': codes,
            'normalizers': [normalizer.__name__ for normalizer in normalizing.normalizers],
            'statistics': statistics(),
            'domains': [
                {
                    'name': domain_name,
                    'metainfo': {
                        'all_package': domain['metainfo_all_package'],
                        'root_section': domain['root_section']
                    }
                }
                for domain_name, domain in datamodel.domains.items()
            ],
            'search_quantities': {
                s.qualified_name: {
                    'name': s.qualified_name,
                    'description': s.description,
                    'many': s.many
                }
                for s in search.search_quantities.values()
                if 'optimade' not in s.qualified_name
            },
            'version': config.meta.version,
            'release': config.meta.release,
            'git': {
                'ref': gitinfo.ref,
                'version': gitinfo.version,
                'commit': gitinfo.commit,
                'log': gitinfo.log
            },
            'oasis': config.keycloak.oasis
        }, 200



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import re
from ase.data import atomic_names, chemical_symbols

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class GamessParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        def get_positions(val):
            # gamess and firefly use diff units
            unit = ureg.bohr if 'BOHR' in val else ureg.angstrom
            res = re.findall(rf'(\w+)\s*{re_f}\s*({re_f} +{re_f} +{re_f})', val)
            labels, positions = [], []
            if res:
                try:
                    labels, positions = list(zip(*res))
                    # firefly prints full name
                    labels = [chemical_symbols[atomic_names.index(s.title())] if len(s) > 2 else s for s in labels]
                    positions = np.array([p.split() for p in positions], dtype=np.dtype(np.float64)) * unit
                except Exception:
                    pass
            return dict(atom_positions=positions, atom_labels=labels)

        super().__init__(
            specifications=dict(
                name='parsers/gamess', code_name='GAMESS',
                code_homepage='https://www.msg.chem.iastate.edu/gamess/versions.html',
                mainfile_contents_re=(
                    r'\s*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\**\s*'
                    r'\s*\*\s*GAMESS VERSION =\s*(.*)\*\s*'
                    r'\s*\*\s*FROM IOWA STATE UNIVERSITY\s*\*\s*')),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            # include code name to distinguish gamess and firefly
            program_version=r'GAMESS VERSION *\= *(.+?) *\*|(Firefly version [\d\.]+)',
            atom_labels_atom_positions=(
                r'(COORDINATES OF ALL ATOMS ARE \(ANGS\)\s*ATOM.+\s*[\s\S]+?)\n *\n|'
                # not sure why \n *\n does not work
                r'(ATOM *ATOMIC *COORDINATES \(BOHR\)\s*CHARGE *X *Y *Z[\s\S]+?)INTER', get_positions),
            energy_total=rf'TOTAL ENERGY \=\s*({re_f})',
            atom_forces=r'GRADIENT \(HARTREE\/BOHR\)([\s\S]+?)\n *\n')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class ElkParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        def get_positions(val):
            labels, positions = [], []
            for line in val.split('\n'):
                species = re.search(r'Species *\:\s*\d+ \((\w+)\)', line)
                if species:
                    label = species.group(1)
                    continue
                position = re.search(rf'\d+\s*\:\s*({re_f} +{re_f} +{re_f})', line)
                if position:
                    labels.append(label)
                    positions.append([float(v) for v in position.group(1).split()])
            return dict(atom_labels=labels, atom_positions_scaled=positions)

        super().__init__(
            specifications=dict(
                name='parsers/elk', code_name='elk', code_homepage='http://elk.sourceforge.net/',
                mainfile_contents_re=r'\| Elk version [0-9.a-zA-Z]+ started \|'),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            program_version=r'Elk version ([\d\.]+) started',
            lattice_vectors=r'Lattice vectors\s*\:\s*([\s\S]+?)\n *\n',
            atom_labels_atom_positions_scaled=(
                r'(Species *\: *\d+ [\s\S]+?Total number of atoms per unit cell.+)',
                get_positions),
            # these are scf values but we write it on scc because we dont want to
            # further complicate the parser by adding a wrapper function
            energy_reference_fermi=(rf'\n *Fermi\s*\:\s*({re_f})', lambda x: [x]),
            energy_total=rf'\n *total energy\s*\:\s*({re_f})',
            energy_sum_eigenvalues=rf'\n *sum of eigenvalues\s*\:\s*({re_f})',
            electronic_kinetic_energy=rf'\n *electron kinetic\s*\:\s*({re_f})',
            energy_correction_hartree=rf'\n *Hartree\s*\:\s*({re_f})',
            energy_C=rf'\n *correlation\s*\:\s*({re_f})',
            energy_X=rf'\n *exchange\s*\:\s*({re_f})')
